// $Header: /cvs/gnusto/src/xpcom/engine/gnusto-engine.js,v 1.79 2004/01/03 03:46:24 marnanel Exp $
// Copyright (c) 2004 Thomas Thurman
// thomas@thurman.org.uk
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of version 2 of the GNU General Public License
// as published by the Free Software Foundation.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have be able to view the GNU General Public License at 
// http://www.gnu.org/copyleft/gpl.html ; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//
const RGTP_CLIENT_CID   = Components.ID("{8b1a4384-1e1e-450b-ba22-e89cd0758616}");
const RGTP_CLIENT_CTRID = "@marnanel.org/archangel/client;1";

function RGTPComponent() { }

RGTPComponent.prototype = {

  receivingFile: false,
  lastMessage: 'none',
  accessLevel: 0,
  chosenEntry: 0,
  currentSubject: '',
  currentParent: '',
  currentChild: '',
  currentItem: '',
  indexDigest: 0,
  indexKeys: 0,
  entries: [],

  connect: function rgtpcl_connect(host, port, username, sharedsecret) {

    this.receivingFile = false;
    this.host = host;
    this.port = port;

    try {

      // Copied wholesale from Chatzilla.

      var sockServiceClass =
	Components.classesByID["{c07e81e0-ef12-11d2-92b6-00105a1b0d64}"];
      
      if (!sockServiceClass)
	throw ("Couldn't get socket service class.");
      
      var sockService = sockServiceClass.getService();
      if (!sockService)
	throw ("Couldn't get socket service.");
  
      this.sockService = sockService.QueryInterface
	(Components.interfaces.nsISocketTransportService);

      if ("nsIStreamProvider" in Components.interfaces) {

	this.transport = this.sockService.createTransport (host, port, null,
                                                              0, 0);

      } else {
	/* use new necko interfaces */
	this.transport = this.sockService.createTransport(null, 0, host, port,
							   null);
      } 

      if (!this.transport) {
	throw ("Error creating transport.");
      }
 
      var openFlags;
      openFlags = 0;
 
      /* no limit on the output stream buffer */
      this.uplink = this.transport.openOutputStream(openFlags, 4096, -1);
      if (!this.uplink)
	throw "Error getting output stream.";
      
      this.downlink = Components.classes['@mozilla.org/scriptableinputstream;1'].
	createInstance().
	QueryInterface(Components.interfaces.nsIScriptableInputStream);
      this.downlink.init(this.transport.openInputStream(openFlags, 1, 1));

      this.setAccessLevelByCode(this.oneRGTPLine());
      
      if (this.access_level==0) {
	this.send('USER guest');
	this.setAccessLevelByCode(this.oneRGTPLine());
      }

      this.indexDigest = 0;

    } catch(ex) {
      dump('EXCEPTION: ');
      dump(ex);
    }
  },

  setAccessLevelByCode: function rgtpcl_setaccesslevelbycode(a_level) {

    if (a_level=='230') {
      this.access_level = 0;
    } else if (a_level=='231') {
      this.access_level = 1;
    } else if (a_level=='232') {
      this.access_level = 2;
    } else if (a_level=='233') {
      this.access_level = 3;
    } else {
      throw "expected access level code!";
    }
  },

  get itemCount() {
    this.syncKeys();
    return this.indexKeys.length;
  },

  loadItemByID: function(itemid) {
    if (this.indexDigest==0) { this.syncIndex(); }
    this.currentItem = itemid;
    this.entries = 0;
  },

  loadItemByNumber: function(number) {
    this.syncKeys();
    this.loadItemByID(this.indexKeys[number]);
  },

  loadMOTD: function() {
    this.send('MOTD');
    this.eatFile(false);
  },

  editItem: function(itemid) {
  },

  loadEntry: function(which) {
    if (this.entries==0) {
      this.syncEntries();
    }

    this.currentEntry = which;
  },

  eatFile: function rgtpcl_eatFile(hasStatus) {

    this.currentSubject = 'no subject?';
    this.currentParent = '';
    this.currentChild = '';
    
    var tempEntry = 0;
    this.entries = [];

    while(1) {
      var code = this.oneRGTPLine();
      if (code=='250') {

	// that's a start
	if (hasStatus) {
	  if (this.oneRGTPLine()!='FILE') throw 'missing header!';
	}

      } else if (code=='STOP') {
	break;
      } else if (code=='FILE' && tempEntry) {
	if (this.lastMessage.substring(0,9)=='Subject: ') {
	  this.currentSubject = this.lastMessage.substring(9);
	}

	tempEntry['text'] = tempEntry['text'] + this.lastMessage + '\n';
      } else {
	// COOKIE, and files that begin with a cookie not marked as such
	// (like the MOTD).

	if (tempEntry) {
	  this.entries.push(tempEntry);
	}

	// FIXME: security hole; I'd rather not eval arbitrary data
	tempEntry = {
	  'seq':  eval('0x'+this.lastMessage.substring(0,8)),
	  'date': eval('0x'+this.lastMessage.substring(9)),
	  'text': '',
	};
      }
    }

    this.entries.push(tempEntry);
  },

  send: function rgtpcl_send(message) {
    message = message + '\r\n';
    try {
      this.uplink.write(message, message.length);
    } catch(ex) {
      dump('Sending: EXCEPTION: ');
      dump(ex);
    }
  },
  
  syncIndex: function rgtpcl_syncIndex() {
    var cacheFile = this.server_cache();

    if (cacheFile.exists()) {

      var fc = Components.classes["@mozilla.org/network/file-input-stream;1"].
	createInstance(Components.interfaces.nsIFileInputStream);

      fc.init(cacheFile, 1, 0, 0);
				
      var sis = new Components.
	Constructor("@mozilla.org/scriptableinputstream;1",
		    "nsIScriptableInputStream")();
      sis.init(fc);

      this.indexDigest = sis.read(cacheFile.fileSize);
      this.indexDigest = eval(this.indexDigest);
      sis.close();
      fc.close();

      // FIXME: sanity check for required properties here.
      // if not, trash the cache.
      // (same if loading from cache caused exception)

    } else {
      this.indexDigest = { 'meta': { 'maxSeq': 0} };
    }

    var showingStatusMessages = 0;

    if (this.indexDigest.meta.maxSeq==0) {
      this.send('INDX');
      showingStatusMessages = 1;
    } else {
      this.send('INDX #'+this.eightHex(this.indexDigest.meta.maxSeq+1));
    }

    if (this.oneRGTPLine()!='250') {
      throw "index retreival failed";
    }

    var notified_datestamp = 0;
    var earliest_datestamp = undefined;
    var latest_datestamp = (new Date().getTime())/1000;
    const notification_period = 2592000; // seconds-- about a month

    if (showingStatusMessages) {
      dump('*******************************************\n');
      dump('**                                       **\n');
      dump('**  Downloading and building the index.  **\n');
      dump('**   This may take a few minutes, but    **\n');
      dump('**     it only needs to happen once.     **\n');
      dump('**                                       **\n');
      dump('*******************************************\n\n');
    }

    while(1) {
      var code = this.oneRGTPLine();
      if (code=='STOP') {
	break;
      } else if (code=='FILE') {
	var indexLine = this.lastMessage;
	var typeCode =  indexLine[103];
	var sequence =  eval('0x'+indexLine.substring(  0,  8));
	var datestamp = eval('0x'+indexLine.substring(  9, 17));
	var itemID =    indexLine.substring( 18, 26);
	var author =    indexLine.substring( 27,101).replace(/\s*$/, '');
	var subject =   indexLine.substring(105    ).replace(/\s*$/, '');

	if (showingStatusMessages && (datestamp-notified_datestamp > notification_period)) {
	  // In lieu of a proper "amount through" dialogue...

	  notified_datestamp = datestamp;

	  if (earliest_datestamp==undefined) {
	    earliest_datestamp = datestamp;
	  }

	  dump('** Reading index: ');
	  dump(Math.round((datestamp-earliest_datestamp)/(latest_datestamp-earliest_datestamp)*100));
	  dump('% **\n');
	}

	if (sequence > this.indexDigest.meta.maxSeq) {
	  this.indexDigest.meta.maxSeq = sequence;
	}

	if (typeCode=='M') {
	  itemID = 'motd';
	}

	if (typeof this.indexDigest[itemID]=='undefined') {
	  this.indexDigest[itemID] = {'count': 0,
			    'subject': subject,
			    'latestBy': '',
			    'latestSeqKnown': 0,
			    'latestSeqRead': 0,
			    'latestDatestamp': 0};
	}

	if (typeCode=='I' || typeCode=='R' || typeCode=='C' || typeCode=='M') {
	  this.indexDigest[itemID].count = this.indexDigest[itemID].count + 1;
	  this.indexDigest[itemID].latestBy = author;

	  if (sequence > this.indexDigest[itemID].latestSeqKnown) {
	    this.indexDigest[itemID].latestSeqKnown = sequence;
	  }

	  if (datestamp > this.indexDigest[itemID].latestDatestamp) {
	    this.indexDigest[itemID].latestDatestamp = datestamp;
	  }
	}

	if (typeCode=='C') {
	  this.indexDigest.meta.newChild = itemID;
	}

	if (typeCode=='F') {
	  var kid = this.indexDigest.meta.newChild;
	  this.indexDigest[itemID].child = kid;
	  this.indexDigest[kid].parent = itemID;
	  this.indexDigest.meta.newChild = null;
	}
	
      } else {
	throw "weird stuff in the index";
      }
    }

    this.saveDigest();
  },

  saveDigest: function rgtpcl_saveDigest() {

    var cache = new Components.
      Constructor('@mozilla.org/network/file-output-stream;1',
		  'nsIFileOutputStream',
		  'init')(this.server_cache(),
			  0x2A, 0600, 0);

    var pickled = uneval(this.indexDigest);
    cache.write(pickled, pickled.length);
    cache.close();
  },

  syncKeys: function rgtpcl_syncKeys() {

    if (this.indexKeys!=0) return;

    this.syncIndex();
    this.indexKeys = [];

    for (k in this.indexDigest) {
      if (k!='meta') {
	this.indexKeys.push(k);
      }
    }
  },

  syncEntries: function rgtpcl_syncEntries() {
    this.send('ITEM '+this.currentItem);
    this.eatFile(true);
  },

  server_cache: function rgtpcl_server_cache() {

    var cache = Components.classes['@mozilla.org/file/directory_service;1'].
      getService(Components.interfaces.nsIProperties).
      get("ProfD", Components.interfaces.nsIFile);

    cache.append('archangel');
    if (!cache.exists()) { cache.create(1, 0700); }
    
    cache.append('cache');
    if (!cache.exists()) { cache.create(1, 0700); }

    cache.append(String(this.host)+'_'+String(this.port));

    return cache;
  },

  eightHex: function rgtpcl_eightHex(number) {
    var temp = number.toString(16);
    while (temp.length<8) {
      temp = '0'+temp;
    }
    return temp;
  },

  oneRGTPLine: function rgtpcl_oneRGTPline() {
    var resultCode = '';
    var receivingNumber = true;
    this.lastMessage = '';

    if (this.receivingFile) {
      receivingNumber = false;
      resultCode = 'FILE';
    }

    while (1) {
      var c = this.downlink.read(1);
      if (c!='' && c!='\r') {
	if (c=='\n') {
	  break;
	} else if (receivingNumber) {
	  if (c==' ') {
	    receivingNumber = false;
	  } else {
	    resultCode = resultCode + c;
	  }
	} else {
	  this.lastMessage = this.lastMessage + c;
	}
      }
    }

    // Time to toggle the file status?

    if (this.receivingFile) {
      if (this.lastMessage=='.') {
	this.receivingFile = false;
	resultCode = 'STOP';
      } else if (this.lastMessage[0]=='.') {
	this.lastMessage = this.lastMessage.substring(1);
      } else if (this.lastMessage[0]=='^') {
	this.lastMessage = this.lastMessage.substring(1);
	resultCode = 'COOKIE';
      }
    } else {
      if (resultCode=='250') {
	this.receivingFile = true;
      }
    }

    return resultCode;
  },

  markItemAsRead: function rgtpcl_markItemAsRead() {
    this.indexDigest[this.currentItem].latestSeqRead =
      this.indexDigest[this.currentItem].latestSeqKnown;
    this.saveDigest();
  },

  get itemID() { return this.currentItem; },

  get level() { dump("level get\n"); return 177; },

  get connected() { /*fixme*/ return true; },

  get entryCount() {
    // if no entries loaded...
    return this.indexDigest[this.currentItem].count;
  },

  get datestamp() { return this.entries[this.currentEntry].date; },
  get seq() { return this.entries[this.currentEntry].seq; },

  get rgtpMessage() { return this.lastMessage; },

  get latestSeqKnown() { return this.indexDigest[this.currentItem].latestSeqKnown; },
  get latestSeqRead() { return this.indexDigest[this.currentItem].latestSeqRead; },
  get latestBy() { return this.indexDigest[this.currentItem].latestBy; },
  get subject() { return this.indexDigest[this.currentItem].subject; },
  get parent() { return this.indexDigest[this.currentItem].parent; },
  get child() { return this.indexDigest[this.currentItem].child; },

  get body() {
    return this.entries[this.currentEntry].text;
  },

  set body(n) { dump("body set to "+String(n)+"\n"); },

  QueryInterface: function(iid) {
    if (iid.equals(Components.interfaces.nsIReverseGossip)) {
      return this;
    } else {
      throw Components.results.NS_ERROR_NO_INTERFACE;
    }
  }
};

////////////////////////////////////////////////////////////////
//
// The factories:
//
// RGTP client handler factory.
//

var RGTPClientHandlerFactory = new Object();

RGTPClientHandlerFactory.createInstance =
function rgtp_clhf_create(outer, iid)
{
  if (outer != null) {
    dump("Don't squish us!\n");
    throw Components.results.NS_ERROR_NO_AGGREGATION;
  }

  if (!iid.equals(Components.interfaces.nsIReverseGossip) &&
      !iid.equals(Components.interfaces.nsISupports)) {
    dump("Ugh, weird interface.\n");
    throw Components.results.NS_ERROR_INVALID_ARG;
  }
  
  return new RGTPComponent();
}

////////////////////////////////////////////////////////////////
//
// The Module
//

var Module = new Object();

Module.registerSelf =
function mod_regself(compMgr, fileSpec, location, type) {
  reg = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
  reg.registerFactoryLocation(RGTP_CLIENT_CID,
			      "RGTP client component",
			      RGTP_CLIENT_CTRID,
			      fileSpec,
			      location,
			      type);
}

Module.getClassObject =
function mod_getclassobj(compMgr, cid, iid) {
  if (cid.equals(RGTP_CLIENT_CID)) {
    return RGTPClientHandlerFactory;
  }

  // okay, so something's weird. give up.
  
  if (!iid.equals(Components.interfaces.nsIFactory))
    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
  
  throw Components.results.NS_ERROR_NO_INTERFACE;
  
}

Module.canUnload =
function mod_canunload(compMgr) {
  return true;
}

////////////////////////////////////////////////////////////////
//
// NSGetModule
//
// It all starts here!

function NSGetModule(compMgr, fileSpec) {
  return Module;
}
